package cd.casic.ci.process.engine.manager.impl;

import cd.casic.ci.common.pipeline.annotation.Plugin;
import cd.casic.ci.process.engine.message.TaskRunMessage;
import cd.casic.ci.process.engine.worker.BaseWorker;
import cd.casic.ci.process.process.dataObject.task.PipTask;
import cd.casic.framework.mq.redis.core.stream.AbstractRedisStreamMessageListener;
import jakarta.annotation.PostConstruct;
import jakarta.annotation.Resource;
import jodd.util.StringUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.context.annotation.ClassPathScanningCandidateComponentProvider;
import org.springframework.core.annotation.AnnotatedElementUtils;
import org.springframework.core.type.filter.AnnotationTypeFilter;
import org.springframework.scheduling.concurrent.ThreadPoolTaskExecutor;
import org.springframework.stereotype.Component;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

@Component
@Slf4j
public class DefaultWorkerManager extends AbstractRedisStreamMessageListener<TaskRunMessage> implements ApplicationContextAware {
    private static final String basePackage = "cd.casic.ci.process.engine.worker";
    private Set<BeanDefinition> candidates;
    private ApplicationContext applicationContext;
    private Map<String,BaseWorker> taskTypeWorkerMap = null;
    @Resource
    private ThreadPoolTaskExecutor workerExecutor;

    private void setTaskTypeWorker(Map<String, BaseWorker> taskTypeWorker) {
        this.taskTypeWorkerMap = taskTypeWorker;
    }

    @PostConstruct
    public void init(){
        ClassPathScanningCandidateComponentProvider provider =
                new ClassPathScanningCandidateComponentProvider(false);
        provider.addIncludeFilter(new AnnotationTypeFilter(Plugin.class));
        candidates = provider.findCandidateComponents(basePackage);
        taskTypeWorkerMap = new HashMap<>(candidates.size());
        for (BeanDefinition candidate : candidates) {
            String beanClassName = candidate.getBeanClassName();
            Class<?> workerClass = null;
            try {
                workerClass = Class.forName(beanClassName);
            } catch (ClassNotFoundException e) {
                continue;
            }
            Object bean = applicationContext.getBean(workerClass);
            if (bean instanceof BaseWorker worker) {
                Plugin annotation = worker.getClass().getAnnotation(Plugin.class);
                String taskType = annotation.taskType();
                if (StringUtils.isNotEmpty(taskType)) {
                    taskTypeWorkerMap.put(taskType,worker);
                }
            }
        }
        log.info("================WorkerManager初始化完毕");
    }
    @Override
    public void onMessage(TaskRunMessage message) {
        log.info("===============接收到消息================");
        try {
            PipTask task = message.getTask();
            String taskType = task.getTaskType();
            BaseWorker baseWorker = taskTypeWorkerMap.get(taskType);
            baseWorker.setContextKey(task.getId());
            workerExecutor.execute(baseWorker);
        }catch (Exception e){
            // TODO 后期可以考虑专门整一个队列
        }
    }

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        this.applicationContext = applicationContext;
    }
}
