package cd.casic.ci.process.engine.runContext;

import cd.casic.ci.process.engine.enums.ContextStateEnum;
import cd.casic.ci.process.process.dataObject.base.PipBaseElement;
import cd.casic.framework.commons.exception.ServiceException;
import cd.casic.framework.commons.exception.enums.GlobalErrorCodeConstants;
import lombok.Data;

import java.time.LocalDateTime;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;
@Data
public abstract class BaseRunContext {
    /**
     *  当前上下文的定义
     * */
    private PipBaseElement contextDef;
    private BaseRunContext parentContext;
    /**
     *  运行状态
     * */
    private AtomicInteger state;
    /**
     *  启动时间
     * */
    private LocalDateTime startTime;
    /**
     *  结束时间
     * */
    private LocalDateTime endTime;
    private String resourceId;
    private String targetId;
    private String targetType;
    /**
     *  整个流水线全局的变量
     * */
    private Map<String,Object> globalVariables;
    /**
     *  当前上下文局部变量
     * */
    private Map<String,Object> localVariables;
    private Map<String,BaseRunContext> childContext;

    public BaseRunContext(PipBaseElement contextDef,BaseRunContext parentContext, LocalDateTime startTime, String resourceId, String targetId, String targetType, Map<String, Object> globalVariables, Map<String, Object> localVariables, Map<String, BaseRunContext> childContext) {
        this.contextDef = contextDef;
        this.parentContext = parentContext;
        this.state = new AtomicInteger(ContextStateEnum.INIT.getCode());
        this.startTime = startTime;
        this.resourceId = resourceId;
        this.targetId = targetId;
        this.targetType = targetType;
        this.globalVariables = globalVariables;
        this.localVariables = localVariables;
        this.childContext = childContext;
    }
    public void changeContextState(ContextStateEnum stateEnum){
        ContextStateEnum curr = ContextStateEnum.getByCode(state.get());
        if (ContextStateEnum.canGoto(curr,stateEnum)) {
            state.compareAndExchange(curr.getCode(),stateEnum.getCode());
            callParentChange(stateEnum);
        }
    }

    private void setState(AtomicInteger state) {
        this.state = state;
    }

    /**
     *  获取当前或者子上下文
     * */
    public abstract BaseRunContext getRunContext(String key);
    public abstract void putChildRunContext(String key,BaseRunContext context);
    public void callParentChange(ContextStateEnum state){
        if (parentContext==null) {
            return;
        }
        if (ContextStateEnum.HAPPY_ENDING.equals(state)||ContextStateEnum.BAD_ENDING.equals(state)) {
            parentContext.checkChildEnd();
        } else if(ContextStateEnum.READY.equals(state)){
            parentContext.checkChildReady();
        } else if(ContextStateEnum.RUNNING.equals(state)){
            parentContext.checkChildRunning();
        }
        parentContext.callParentChange(state);
    }
    /**
     *  查找子类是否全部完成，如果子类全部完成则父类也完成
     * */
    public void checkChildEnd() throws ServiceException{
        int result = ContextStateEnum.HAPPY_ENDING.getCode();
        for (Map.Entry<String, BaseRunContext> entry : childContext.entrySet()) {
            BaseRunContext child = entry.getValue();
            int state = child.getState().get();
            if (!ContextStateEnum.HAPPY_ENDING.getCode().equals(state)&&!ContextStateEnum.BAD_ENDING.getCode().equals(state)) {
                return;
            }
            result&=state;
        }
        boolean end = false;
        if (ContextStateEnum.HAPPY_ENDING.getCode()==result) {
            if (ContextStateEnum.canGoto(ContextStateEnum.getByCode(state.get()),ContextStateEnum.HAPPY_ENDING)) {
                this.state.compareAndExchange(state.get(),ContextStateEnum.HAPPY_ENDING.getCode());
                end = true;
            }
        } else {
            if (ContextStateEnum.canGoto(ContextStateEnum.getByCode(state.get()),ContextStateEnum.BAD_ENDING)) {
                this.state.compareAndExchange(state.get(),ContextStateEnum.BAD_ENDING.getCode());
                end = true;
            }
        }
        if (!end) {
            throw new ServiceException(GlobalErrorCodeConstants.PIPELINE_ERROR.getCode(),"状态有误");
        }
    }
    /**
     *  查找子类是否全部就绪，如果子类全部完成则父类也就绪
     * */
    public void checkChildReady() throws ServiceException{
        int result = ContextStateEnum.READY.getCode();
        for (Map.Entry<String, BaseRunContext> entry : childContext.entrySet()) {
            BaseRunContext child = entry.getValue();
            int state = child.getState().get();
            if (!ContextStateEnum.READY.getCode().equals(state)) {
                return;
            }
            result&=state;
        }
        boolean ready = false;
        if (ContextStateEnum.READY.getCode()==result) {
            if (ContextStateEnum.canGoto(ContextStateEnum.getByCode(state.get()),ContextStateEnum.READY)) {
                this.state.compareAndExchange(state.get(),ContextStateEnum.READY.getCode());
                ready = true;
            }
        }
        if (!ready) {
            throw new ServiceException(GlobalErrorCodeConstants.PIPELINE_ERROR.getCode(),"状态有误");
        }
    }
    /**
     *  查找子类是否存在开始运行的，如果有则父状态变成running
     * */
    public void checkChildRunning() throws ServiceException{
        Boolean runningFlag = false;
        for (Map.Entry<String, BaseRunContext> entry : childContext.entrySet()) {
            BaseRunContext child = entry.getValue();
            int state = child.getState().get();
            if (ContextStateEnum.READY.getCode().equals(state)) {
                runningFlag = true;
                break;
            }
        }
        if (runningFlag) {
            if (ContextStateEnum.canGoto(ContextStateEnum.getByCode(state.get()),ContextStateEnum.RUNNING)) {
                this.state.compareAndExchange(state.get(),ContextStateEnum.RUNNING.getCode());
            } else{
                throw new ServiceException(GlobalErrorCodeConstants.PIPELINE_ERROR.getCode(),"状态有误");
            }
        }
    }
}
